"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const chalk_1 = require("chalk");
const debug_1 = __importDefault(require("debug"));
const deepmerge_1 = __importDefault(require("deepmerge"));
const log_symbols_1 = require("log-symbols");
const node_util_1 = require("node:util");
const ora_1 = __importDefault(require("ora"));
const pluralize_1 = __importDefault(require("pluralize"));
const helpers_1 = require("yargs/helpers");
const yargs_1 = __importDefault(require("yargs/yargs"));
const config_file_1 = require("./config-file");
const error_1 = require("./error");
const event_1 = require("./event");
const options_1 = require("./options");
const smoker_1 = require("./smoker");
const util_1 = require("./util");
const debug = (0, debug_1.default)('midnight-smoker:cli');
const BEHAVIOR_GROUP = 'Behavior:';
/**
 * Main entry point for the CLI script
 * @param args Raw command-line arguments array
 */
async function main(args) {
    const y = (0, yargs_1.default)(args);
    const result = await (0, util_1.readPackageJson)({ cwd: __dirname });
    if (!result) {
        throw new Error('midnight-smoker could not find its own package.json! bogus');
    }
    const { packageJson } = result;
    const { version, homepage } = packageJson;
    let verbose = false;
    try {
        const config = await (0, config_file_1.readConfigFile)();
        await y
            .scriptName('smoker')
            .command('* [scripts..]', 'Run tests against a package as it would be published', (yargs) => {
            /**
             * Reusable config for array-type options
             */
            const arrayOptConfig = {
                requiresArg: true,
                array: true,
                type: 'string',
            };
            return yargs
                .positional('scripts', {
                describe: 'Script(s) in package.json to run',
                type: 'string',
            })
                .options({
                add: {
                    describe: 'Additional dependency to provide to script(s)',
                    group: BEHAVIOR_GROUP,
                    ...arrayOptConfig,
                },
                all: {
                    describe: 'Run script in all workspaces',
                    group: BEHAVIOR_GROUP,
                    type: 'boolean',
                },
                bail: {
                    describe: 'When running scripts, halt on first error',
                    group: BEHAVIOR_GROUP,
                    type: 'boolean',
                },
                'include-root': {
                    describe: "Include the workspace root; must provide '--all'",
                    group: BEHAVIOR_GROUP,
                    implies: 'all',
                    type: 'boolean',
                },
                json: {
                    describe: 'Output JSON only',
                    group: BEHAVIOR_GROUP,
                    type: 'boolean',
                },
                linger: {
                    describe: 'Do not clean up temp dir(s) after completion',
                    group: BEHAVIOR_GROUP,
                    hidden: true,
                    type: 'boolean',
                },
                pm: {
                    describe: 'Run script(s) with a specific package manager; <npm|yarn|pnpm>[@version]',
                    group: BEHAVIOR_GROUP,
                    default: [options_1.DEFAULT_PACKAGE_MANAGER_ID],
                    ...arrayOptConfig,
                },
                loose: {
                    describe: 'Ignore missing scripts (used with --all)',
                    type: 'boolean',
                    group: BEHAVIOR_GROUP,
                    implies: 'all',
                },
                workspace: {
                    describe: 'Run script in a specific workspace or workspaces',
                    group: BEHAVIOR_GROUP,
                    ...arrayOptConfig,
                },
                checks: {
                    describe: 'Run built-in checks',
                    group: BEHAVIOR_GROUP,
                    type: 'boolean',
                },
            });
        }, async (argv) => {
            const opts = (0, deepmerge_1.default)(config, argv);
            if (opts.pm?.some((pm) => pm.startsWith('pnpm'))) {
                throw new error_1.NotImplementedError('pnpm is currently unsupported');
            }
            verbose = Boolean(opts.verbose || debug_1.default.enabled('midnight-smoker'));
            if (verbose) {
                debug_1.default.enable('midnight-smoker');
            }
            debug('Final options: %O', opts);
            const smoker = await smoker_1.Smoker.init(opts);
            if (opts.json) {
                // if we don't have any scripts, the script-related ones will remain null
                let totalPackages = null;
                let totalScripts = null;
                let totalPackageManagers = null;
                let failedScripts = null;
                let passedScripts = null;
                let totalChecks = null;
                let failedChecks = null;
                let passedChecks = null;
                smoker
                    .once(event_1.Event.INSTALL_BEGIN, ({ uniquePkgs, packageManagers }) => {
                    totalPackages = uniquePkgs.length;
                    totalPackageManagers = packageManagers.length;
                })
                    .once(event_1.Event.INSTALL_FAILED, () => {
                    process.exitCode = 1;
                })
                    .once(event_1.Event.RUN_SCRIPTS_BEGIN, ({ total }) => {
                    totalScripts = total;
                })
                    .once(event_1.Event.RUN_SCRIPTS_FAILED, ({ failed, passed }) => {
                    failedScripts = failed;
                    passedScripts = passed;
                    smoker.removeAllListeners(event_1.Event.RUN_SCRIPTS_OK);
                    process.exitCode = 1;
                })
                    .once(event_1.Event.RUN_SCRIPTS_OK, ({ passed }) => {
                    passedScripts = passed;
                    failedScripts = 0;
                    smoker.removeAllListeners(event_1.Event.RUN_SCRIPTS_FAILED);
                })
                    .once(event_1.Event.RUN_CHECKS_BEGIN, ({ total }) => {
                    totalChecks = total;
                })
                    .once(event_1.Event.RUN_CHECKS_FAILED, ({ failed, passed }) => {
                    failedChecks = failed.length;
                    passedChecks = passed.length;
                    smoker.removeAllListeners(event_1.Event.RUN_CHECKS_OK);
                })
                    .on(event_1.Event.RUN_CHECK_FAILED, ({ config }) => {
                    if (config.severity === 'error') {
                        process.exitCode = 1;
                    }
                })
                    .once(event_1.Event.RUN_CHECKS_OK, ({ passed }) => {
                    passedChecks = passed.length;
                    failedChecks = 0;
                    smoker.removeAllListeners(event_1.Event.RUN_CHECKS_FAILED);
                })
                    .once(event_1.Event.SMOKE_OK, () => {
                    smoker.removeAllListeners();
                })
                    .once(event_1.Event.SMOKE_FAILED, () => {
                    smoker.removeAllListeners();
                });
                let output;
                try {
                    const results = await smoker.smoke();
                    const stats = {
                        totalPackages,
                        totalPackageManagers,
                        totalScripts,
                        failedScripts,
                        passedScripts,
                        totalChecks,
                        failedChecks,
                        passedChecks,
                    };
                    output = { results, stats };
                }
                catch (err) {
                    const stats = {
                        totalPackages,
                        totalPackageManagers,
                        totalScripts,
                        failedScripts,
                        passedScripts,
                        totalChecks,
                        failedChecks,
                        passedChecks,
                    };
                    output = { error: err.message, stats };
                }
                finally {
                    smoker.removeAllListeners();
                }
                console.log(JSON.stringify(output, null, 2));
            }
            else {
                const spinner = (0, ora_1.default)();
                const scriptFailedEvts = [];
                const checkFailedEvts = [];
                smoker
                    .once(event_1.Event.SMOKE_BEGIN, () => {
                    console.error(`💨 ${(0, chalk_1.blue)('midnight-smoker')} ${(0, chalk_1.white)(`v${version}`)}`);
                })
                    .once(event_1.Event.PACK_BEGIN, () => {
                    let what;
                    if (opts.workspace?.length) {
                        what = (0, pluralize_1.default)('workspace', opts.workspace.length, true);
                    }
                    else if (opts.all) {
                        what = 'all workspaces';
                        if (opts.includeRoot) {
                            what += ' (and the workspace root)';
                        }
                    }
                    else {
                        what = 'current project';
                    }
                    spinner.start(`Packing ${what}…`);
                })
                    .once(event_1.Event.PACK_OK, ({ uniquePkgs, packageManagers }) => {
                    let msg = `Packed ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} using `;
                    if (packageManagers.length > 1) {
                        msg += `${(0, pluralize_1.default)('package manager', packageManagers.length, true)}`;
                    }
                    else {
                        msg += `${packageManagers[0]}`;
                    }
                    msg += '…';
                    spinner.succeed(msg);
                })
                    .once(event_1.Event.PACK_FAILED, () => {
                    spinner.fail('Failed while packing!');
                    process.exitCode = 1;
                })
                    .once(event_1.Event.INSTALL_BEGIN, ({ uniquePkgs, packageManagers, additionalDeps }) => {
                    let msg = `Installing ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} from tarball`;
                    if (additionalDeps.length) {
                        msg += ` with ${(0, pluralize_1.default)('additional dependency', additionalDeps.length, true)}`;
                    }
                    if (packageManagers.length > 1) {
                        msg += ` using ${(0, pluralize_1.default)('package manager', packageManagers.length, true)}`;
                    }
                    else {
                        msg += ` using ${packageManagers[0]}`;
                    }
                    msg += '…';
                    spinner.start(msg);
                })
                    .once(event_1.Event.INSTALL_FAILED, () => {
                    spinner.fail('Failed while installing!');
                    process.exitCode = 1;
                })
                    .once(event_1.Event.INSTALL_OK, ({ uniquePkgs }) => {
                    spinner.succeed(`Installed ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} from tarball`);
                })
                    .once(event_1.Event.RUN_CHECKS_BEGIN, ({ total }) => {
                    spinner.start(`Running 0/${total} checks…`);
                })
                    .on(event_1.Event.RUN_CHECK_BEGIN, ({ current, total }) => {
                    spinner.text = `Running check ${current}/${total}…`;
                })
                    .on(event_1.Event.RUN_CHECK_FAILED, (evt) => {
                    checkFailedEvts.push(evt);
                    if (evt.config.severity === 'error') {
                        process.exitCode = 1;
                    }
                })
                    .once(event_1.Event.RUN_CHECKS_OK, ({ total }) => {
                    spinner.succeed(`Successfully ran ${(0, pluralize_1.default)('check', total, true)}`);
                })
                    .once(event_1.Event.RUN_CHECKS_FAILED, ({ total, failed }) => {
                    spinner.fail(`${(0, pluralize_1.default)('check', failed.length, true)} of ${total} failed`);
                    // TODO: move this crunching somewhere else
                    const failedByPackage = checkFailedEvts
                        .map((evt) => evt.failed)
                        .flat()
                        .reduce((acc, failed) => {
                        const pkgName = failed.context.pkgJson.name ?? failed.context.pkgPath;
                        acc[pkgName] = [...(acc[pkgName] ?? []), failed];
                        return acc;
                    }, {});
                    for (const [pkgName, failed] of Object.entries(failedByPackage)) {
                        let text = `Check failures in package ${(0, chalk_1.cyan)(pkgName)}:\n`;
                        for (const { message, severity } of failed) {
                            if (severity === 'error') {
                                text += `» ${log_symbols_1.error} ${(0, chalk_1.red)(message)}\n`;
                            }
                            else {
                                text += `» ${log_symbols_1.warning} ${(0, chalk_1.yellow)(message)}\n`;
                            }
                        }
                        spinner.info(text);
                    }
                })
                    .on(event_1.Event.RULE_ERROR, (err) => {
                    spinner.fail(err.message);
                    process.exitCode = 1;
                })
                    .once(event_1.Event.RUN_SCRIPTS_BEGIN, ({ total }) => {
                    spinner.start(`Running script 0/${total}…`);
                })
                    .on(event_1.Event.RUN_SCRIPT_BEGIN, ({ current, total }) => {
                    spinner.text = `Running script ${current}/${total}…`;
                })
                    .on(event_1.Event.RUN_SCRIPT_FAILED, (evt) => {
                    scriptFailedEvts.push(evt);
                    process.exitCode = 1;
                })
                    .once(event_1.Event.RUN_SCRIPTS_OK, ({ total }) => {
                    spinner.succeed(`Successfully ran ${(0, pluralize_1.default)('script', total, true)}`);
                })
                    .once(event_1.Event.RUN_SCRIPTS_FAILED, ({ total, failed: failures }) => {
                    spinner.fail(`${failures} of ${total} ${(0, pluralize_1.default)('script', total)} failed`);
                    for (const evt of scriptFailedEvts) {
                        spinner.info(`${(0, chalk_1.red)('Script failure')} details for package "${(0, chalk_1.cyan)(evt.pkgName)}":\n» ${(0, chalk_1.yellow)(evt.error)}\n`);
                    }
                    process.exitCode = 1;
                })
                    .once(event_1.Event.SMOKE_FAILED, (err) => {
                    spinner.fail(err?.message ?? err);
                })
                    .once(event_1.Event.SMOKE_OK, () => {
                    spinner.succeed('Lovey-dovey! 💖');
                })
                    .once(event_1.Event.LINGERED, (dirs) => {
                    console.error(`${log_symbols_1.info} Lingering ${(0, pluralize_1.default)('temp directory', dirs.length)}:\n`);
                    for (const dir of dirs) {
                        console.error(`» ${(0, chalk_1.yellow)(dir)}`);
                    }
                });
                try {
                    await smoker.smoke();
                }
                catch (e) {
                    const err = e;
                    console.error();
                    if (verbose) {
                        console.error((0, node_util_1.inspect)(err, { depth: null, colors: true }));
                    }
                    else {
                        console.error(err.message);
                        console.error((0, chalk_1.dim)('(use --verbose for more details)'));
                    }
                    process.exitCode = 1;
                }
            }
        })
            .options({
            verbose: {
                describe: 'Verbose output',
                type: 'boolean',
                global: true,
            },
        })
            .epilog(`For more info, see ${homepage}\n`)
            .showHelpOnFail(false)
            .help()
            .strict()
            .parseAsync();
    }
    catch (err) {
        // I think anything here would happen before we started smoking
        console.error(err);
        process.exitCode = 1;
    }
}
main((0, helpers_1.hideBin)(process.argv));
//# sourceMappingURL=cli.js.map