/**
 * All the errors
 * @module
 */
/// <reference types="node" />
import { StaticCheckContext } from '.';
import { SmokeResults } from './types';
/**
 * Options for {@link SmokerError} with a generic `Cause` type for `cause` prop.
 */
export interface SmokerErrorOpts<Cause extends object | void = void> extends ErrorOptions {
    code?: string;
    cause?: Cause;
}
/**
 * Base class for all exceptions thrown by `midnight-smoker`.
 */
export declare abstract class SmokerError<Cause extends object | void = void> extends Error implements NodeJS.ErrnoException {
    /**
     * Each subclass will have a unique `code`
     */
    readonly code?: string;
    /**
     * Each subclass defines the shape of its `cause`.
     */
    readonly cause?: Cause;
    constructor(message: string, opts?: SmokerErrorOpts<Cause>);
    toJSON(): {
        message: string;
        name: string;
        stack: string | undefined;
        cause: Cause | undefined;
        code: string | undefined;
    };
}
export declare class InvalidArgError extends SmokerError<{
    arg?: string;
}> {
    constructor(message: string, arg?: string);
}
export declare class NotImplementedError extends SmokerError {
    readonly name = "NotImplementedError";
    constructor(message: string);
}
export declare class DirCreationError extends SmokerError<{
    error: NodeJS.ErrnoException;
}> {
    readonly name = "DirCreationError";
    constructor(message: string, error: NodeJS.ErrnoException);
}
export declare class DirDeletionError extends SmokerError<{
    dir: string;
    error: NodeJS.ErrnoException;
}> {
    readonly name = "DirDeletionError";
    constructor(message: string, dir: string, error: NodeJS.ErrnoException);
}
export declare class PackageManagerIdError extends SmokerError<{
    pmId: string;
}> {
    readonly name = "PackageManagerIdError";
    constructor();
}
export declare class PackageManagerError extends SmokerError<{
    error: Error;
    pmId: string;
}> {
    readonly name = "PackageManagerError";
    constructor(message: string, pmId: string, error: Error);
}
export declare class SmokeFailedError extends SmokerError<{
    results?: SmokeResults;
    error?: Error;
}> {
    constructor(message: string, opts?: {
        results?: SmokeResults;
        error?: Error;
    });
}
export declare class MissingPackageJsonError extends SmokerError<{
    cwd: string;
}> {
    constructor(message: string, cwd: string);
}
export declare class UnreadablePackageJsonError extends SmokerError<{
    pkgJsonPath: string;
    error: Error;
}> {
    constructor(message: string, pkgJsonPath: string, error: Error);
}
export declare class RunScriptError extends SmokerError<{
    script: string;
    pkgName: string;
    pm: string;
    error?: Error;
    exitCode?: number;
    output?: string;
}> {
    constructor(message: string, script: string, pkgName: string, pm: string, opts?: {
        error?: Error;
        exitCode?: number;
        output?: string;
    });
    toString(): string;
}
export declare class UnknownScriptError extends SmokerError<{
    script: string;
    pkgName: string;
}> {
    constructor(message: string, script: string, pkgName: string);
}
export type ScriptError = RunScriptError | UnknownScriptError;
export declare class InstallError extends SmokerError<{
    pm: string;
    error?: Error;
    exitCode?: number;
    output?: string;
}> {
    constructor(message: string, pm: string, opts?: {
        error?: Error;
        exitCode?: number;
        output?: string;
    });
}
export declare class PackError extends SmokerError<{
    pm: string;
    error?: Error;
    exitCode?: number;
    output?: string;
}> {
    constructor(message: string, pm: string, opts?: {
        error?: Error;
        exitCode?: number;
        output?: string;
    });
}
export declare class PackParseError extends SmokerError<{
    pm: string;
    error: Error;
    output: string;
}> {
    constructor(message: string, pm: string, error: Error, output: string);
}
export declare class UnknownVersionError extends SmokerError<{
    pm: string;
    version: string;
}> {
    constructor(message: string, pm: string, version: string);
}
export declare class UnknownVersionRangeError extends SmokerError<{
    pm: string;
    versionRange: string;
}> {
    constructor(message: string, pm: string, versionRange: string);
}
export declare class UnknownDistTagError extends SmokerError<{
    pkgName: string;
    tag: string;
}> {
    constructor(message: string, pkgName: string, tag: string);
}
export declare class UnsupportedPackageManagerError extends SmokerError<{
    name: string;
    version: string;
}> {
    constructor(message: string, name: string, version: string);
}
export declare class RuleError extends SmokerError<{
    context: StaticCheckContext;
    ruleName: string;
    error: Error;
}> {
    constructor(message: string, context: StaticCheckContext, ruleName: string, error: Error);
}
//# sourceMappingURL=error.d.ts.map