"use strict";
/**
 * All the errors
 * @module
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.RuleError = exports.UnsupportedPackageManagerError = exports.UnknownDistTagError = exports.UnknownVersionRangeError = exports.UnknownVersionError = exports.PackParseError = exports.PackError = exports.InstallError = exports.UnknownScriptError = exports.RunScriptError = exports.UnreadablePackageJsonError = exports.MissingPackageJsonError = exports.SmokeFailedError = exports.PackageManagerError = exports.PackageManagerIdError = exports.DirDeletionError = exports.DirCreationError = exports.NotImplementedError = exports.InvalidArgError = exports.SmokerError = void 0;
const chalk_1 = require("chalk");
/**
 * Base class for all exceptions thrown by `midnight-smoker`.
 */
class SmokerError extends Error {
    constructor(message, opts) {
        super(message, opts);
        this.code = opts?.code;
    }
    toJSON() {
        return {
            message: this.message,
            name: this.name,
            stack: this.stack,
            cause: this.cause,
            code: this.code,
        };
    }
}
exports.SmokerError = SmokerError;
class InvalidArgError extends SmokerError {
    constructor(message, arg) {
        super(message, { code: 'ESMOKER_INVALIDARG', cause: { arg } });
    }
}
exports.InvalidArgError = InvalidArgError;
class NotImplementedError extends SmokerError {
    constructor(message) {
        super(message, { code: 'ESMOKER_NOTIMPLEMENTED' });
        this.name = 'NotImplementedError';
    }
}
exports.NotImplementedError = NotImplementedError;
class DirCreationError extends SmokerError {
    constructor(message, error) {
        super(message, { code: 'ESMOKER_DIRCREATION', cause: { error } });
        this.name = 'DirCreationError';
    }
}
exports.DirCreationError = DirCreationError;
class DirDeletionError extends SmokerError {
    constructor(message, dir, error) {
        super(message, { code: 'ESMOKER_DIRDELETION', cause: { dir, error } });
        this.name = 'DirDeletionError';
    }
}
exports.DirDeletionError = DirDeletionError;
class PackageManagerIdError extends SmokerError {
    constructor() {
        super('Could not find package manager ID; please report this bug at https://github.com/boneskull/midnight-smoker/issues/new', { code: 'ESMOKER_PACKAGEMANAGERID' });
        this.name = 'PackageManagerIdError';
    }
}
exports.PackageManagerIdError = PackageManagerIdError;
class PackageManagerError extends SmokerError {
    constructor(message, pmId, error) {
        super(message, { code: 'ESMOKER_PACKAGEMANAGER', cause: { pmId, error } });
        this.name = 'PackageManagerError';
    }
}
exports.PackageManagerError = PackageManagerError;
class SmokeFailedError extends SmokerError {
    constructor(message, opts = {}) {
        super(message, { code: 'ESMOKER_SMOKEFAILED', cause: opts });
    }
}
exports.SmokeFailedError = SmokeFailedError;
class MissingPackageJsonError extends SmokerError {
    constructor(message, cwd) {
        super(message, { code: 'ESMOKER_MISSINGPACKAGEJSON', cause: { cwd } });
    }
}
exports.MissingPackageJsonError = MissingPackageJsonError;
class UnreadablePackageJsonError extends SmokerError {
    constructor(message, pkgJsonPath, error) {
        super(message, {
            code: 'ESMOKER_UNREADABLEPACKAGEJSON',
            cause: { pkgJsonPath, error },
        });
    }
}
exports.UnreadablePackageJsonError = UnreadablePackageJsonError;
class RunScriptError extends SmokerError {
    constructor(message, script, pkgName, pm, opts = {}) {
        super(message, {
            code: 'ESMOKER_RUNSCRIPT',
            cause: { script, pkgName, pm, ...opts },
        });
    }
    toString() {
        if (this.cause?.error) {
            return `${this.message}: ${(0, chalk_1.bold)(this.cause.error.message ?? this.cause.error)}`;
        }
        if (this.cause?.output) {
            return `${this.message}: ${(0, chalk_1.bold)(this.cause.output)}`;
        }
        return this.message;
    }
}
exports.RunScriptError = RunScriptError;
class UnknownScriptError extends SmokerError {
    constructor(message, script, pkgName) {
        super(message, { code: 'ESMOKER_UNKNOWNSCRIPT', cause: { script, pkgName } });
    }
}
exports.UnknownScriptError = UnknownScriptError;
class InstallError extends SmokerError {
    constructor(message, pm, opts = {}) {
        super(message, { code: 'ESMOKER_INSTALL', cause: { pm, ...opts } });
    }
}
exports.InstallError = InstallError;
class PackError extends SmokerError {
    constructor(message, pm, opts = {}) {
        super(message, { code: 'ESMOKER_PACK', cause: { pm, ...opts } });
    }
}
exports.PackError = PackError;
class PackParseError extends SmokerError {
    constructor(message, pm, error, output) {
        super(message, {
            code: 'ESMOKER_PACKPARSE',
            cause: { error, pm, output },
        });
    }
}
exports.PackParseError = PackParseError;
class UnknownVersionError extends SmokerError {
    constructor(message, pm, version) {
        super(message, {
            code: 'ESMOKER_UNKNOWNVERSION',
            cause: { pm, version },
        });
    }
}
exports.UnknownVersionError = UnknownVersionError;
class UnknownVersionRangeError extends SmokerError {
    constructor(message, pm, versionRange) {
        super(message, {
            code: 'ESMOKER_UNKNOWNVERSIONRANGE',
            cause: { pm, versionRange },
        });
    }
}
exports.UnknownVersionRangeError = UnknownVersionRangeError;
class UnknownDistTagError extends SmokerError {
    constructor(message, pkgName, tag) {
        super(message, { code: 'ESMOKER_UNKNOWNDISTTAG', cause: { pkgName, tag } });
    }
}
exports.UnknownDistTagError = UnknownDistTagError;
class UnsupportedPackageManagerError extends SmokerError {
    constructor(message, name, version) {
        super(message, {
            code: 'ESMOKER_UNSUPPORTEDPACKAGEMANAGER',
            cause: { name, version },
        });
    }
}
exports.UnsupportedPackageManagerError = UnsupportedPackageManagerError;
class RuleError extends SmokerError {
    constructor(message, context, ruleName, error) {
        super(message, {
            code: 'ESMOKER_RULEERROR',
            cause: { context, ruleName, error },
        });
    }
}
exports.RuleError = RuleError;
//# sourceMappingURL=error.js.map