"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseOptions = exports.zSmokerOptions = exports.zRawSmokerOptions = exports.DEFAULT_PACKAGE_MANAGER_ID = void 0;
/**
 * Handles parsing of options (CLI, API, or config file) for `midnight-smoker`
 * @module
 */
const debug_1 = __importDefault(require("debug"));
const zod_1 = __importDefault(require("zod"));
const rules_1 = require("./rules");
const zod_validation_error_1 = require("zod-validation-error");
const schema_util_1 = require("./schema-util");
const debug = (0, debug_1.default)('midnight-smoker:options');
/**
 * @internal
 */
exports.DEFAULT_PACKAGE_MANAGER_ID = 'npm@latest';
exports.zRawSmokerOptions = zod_1.default
    .object({
    add: schema_util_1.zStringOrArray.describe('Add an extra package to the list of packages to be installed.'),
    all: schema_util_1.zFalse.describe('Operate on all workspaces. The root workspace is omitted unless `includeRoot` is `true`.'),
    bail: schema_util_1.zFalse.describe('Fail on first script failure.'),
    includeRoot: schema_util_1.zFalse.describe('Operate on the root workspace. Only has an effect if `all` is `true`.'),
    json: schema_util_1.zFalse.describe('Output JSON only.'),
    linger: schema_util_1.zFalse.describe('Do not delete temp directories after completion.'),
    verbose: schema_util_1.zFalse.describe('Verbose logging.'),
    workspace: schema_util_1.zStringOrArray.describe('One or more workspaces to run scripts in.'),
    pm: schema_util_1.zStringOrArray
        .default(exports.DEFAULT_PACKAGE_MANAGER_ID)
        .describe('Package manager(s) to use.'),
    script: schema_util_1.zStringOrArray.describe('Script(s) to run. Alias of `scripts`.'),
    scripts: schema_util_1.zStringOrArray.describe('Script(s) to run. Alias of `script`.'),
    loose: schema_util_1.zFalse.describe('If `true`, fail if a workspace is missing a script.'),
    checks: schema_util_1.zTrue.describe('If `false`, run no builtin checks.'),
    rules: rules_1.zCheckOptions,
})
    .describe('midnight-smoker options schema');
/**
 * @internal
 */
exports.zSmokerOptions = exports.zRawSmokerOptions
    .transform((cfg, ctx) => {
    if (cfg.all) {
        if (cfg.workspace.length) {
            ctx.addIssue({
                code: zod_1.default.ZodIssueCode.custom,
                message: 'Option "workspace" is mutually exclusive with "all"',
            });
            return zod_1.default.NEVER;
        }
    }
    else if (cfg.loose) {
        ctx.addIssue({
            code: zod_1.default.ZodIssueCode.custom,
            message: 'Option "loose" requires "all" to be set',
        });
        return zod_1.default.NEVER;
    }
    // stuff `scripts` into into `script`
    const script = [...new Set([...cfg.script, ...cfg.scripts])];
    return {
        ...cfg,
        script,
    };
})
    .brand()
    .readonly();
/**
 * Parses options for `Smoker`.
 *
 * @param opts Options for `Smoker`. Could come from CLI, config file, API, or some combination thereof.
 * @returns Parsed & normalized options.
 */
function parseOptions(opts) {
    if (parseOptions.cache.has(opts)) {
        return parseOptions.cache.get(opts);
    }
    let result;
    try {
        result = exports.zSmokerOptions.parse(opts ?? {});
    }
    catch (err) {
        throw (0, zod_validation_error_1.fromZodError)(err);
    }
    parseOptions.cache.set(opts, result);
    parseOptions.cache.set(result, result);
    debug('Normalized options: %O', result);
    return result;
}
exports.parseOptions = parseOptions;
parseOptions.cache = new Map();
//# sourceMappingURL=options.js.map