"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GenericNpmPackageManager = void 0;
const error_1 = require("../error");
class GenericNpmPackageManager {
    constructor(executor, opts) {
        this.executor = executor;
        this.opts = opts;
        this.name = 'npm';
    }
    async runScript(manifest) {
        if (!manifest) {
            throw new TypeError('(runScript) "manifest" arg is required');
        }
        const { script, packedPkg } = manifest;
        const args = ['run', script];
        const { pkgName, installPath: cwd } = packedPkg;
        let result;
        try {
            const rawResult = await this.executor.exec(args, {
                cwd,
            });
            result = { pkgName, script, rawResult, cwd };
        }
        catch (err) {
            const error = err;
            result = {
                pkgName,
                script,
                rawResult: error,
                cwd,
            };
            if (this.opts.loose && /missing script:/i.test(error.stderr)) {
                result.skipped = true;
            }
            else {
                result.error = new error_1.RunScriptError(`Script "${script}" in package "${pkgName}" failed`, script, pkgName, this.name, { error, exitCode: error.exitCode, output: error.stderr });
            }
        }
        if (!result.error && !result.skipped && result.rawResult.failed) {
            if (result.rawResult.stderr &&
                /missing script:/i.test(result.rawResult.stderr)) {
                if (!this.opts.loose) {
                    result.error = new error_1.UnknownScriptError(`Script "${script}" in package "${pkgName}" not found`, script, pkgName);
                }
                else {
                    result.skipped = true;
                }
            }
            else {
                let message;
                if (result.rawResult.exitCode) {
                    message = `Script "${script}" in package "${pkgName}" failed with exit code ${result.rawResult.exitCode}`;
                }
                else {
                    message = `Script "${script}" in package "${pkgName}" failed`;
                }
                result.error = new error_1.RunScriptError(message, script, pkgName, this.name, {
                    exitCode: result.rawResult.exitCode,
                    output: result.rawResult.all,
                });
            }
        }
        if (result.error) {
            this.debug(`(runScripts) Script "%s" in package "%s" failed; continuing...`, script, pkgName);
        }
        else {
            this.debug('(runScripts) Successfully executed script %s in package %s', script, pkgName);
        }
        return result;
    }
}
exports.GenericNpmPackageManager = GenericNpmPackageManager;
//# sourceMappingURL=npm.js.map