"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Npm7 = void 0;
const chalk_1 = require("chalk");
const debug_1 = __importDefault(require("debug"));
const node_path_1 = __importDefault(require("node:path"));
const error_1 = require("../error");
const npm_1 = require("./npm");
class Npm7 extends npm_1.GenericNpmPackageManager {
    constructor(executor, opts = {}) {
        super(executor, opts);
        this.name = 'npm';
        this.debug = (0, debug_1.default)(`midnight-smoker:pm:npm7`);
    }
    static accepts(semver) {
        return Boolean(semver.compare('7.0.0') - semver.compare('9.0.0'));
    }
    static load(executor, opts) {
        return new Npm7(executor, opts);
    }
    async install(manifest) {
        const { packedPkgs, tarballRootDir } = manifest;
        if (!packedPkgs?.length) {
            throw new TypeError('(install) Non-empty "manifest" arg is required');
        }
        const additionalDeps = manifest.additionalDeps ?? [];
        // otherwise we get a deprecation warning
        const installArgs = [
            'install',
            '--no-package-lock',
            '--global-style',
            ...packedPkgs.map(({ tarballFilepath }) => tarballFilepath),
            ...additionalDeps,
        ];
        let installResult;
        try {
            installResult = await this.executor.exec(installArgs, {
                cwd: tarballRootDir,
            });
        }
        catch (e) {
            const err = e;
            let parsed;
            try {
                this.debug('Trying to parse stdout: %s', err.stdout);
                parsed = JSON.parse(err.stdout);
            }
            catch {
                // ignore
            }
            if (parsed?.error) {
                throw new error_1.InstallError(`Package manager "${this.name}" failed to install packages:\n\n${(0, chalk_1.red)(parsed.error.summary)}`, this.name, {
                    error: err,
                    output: err.stderr,
                });
            }
            throw new error_1.InstallError(`Package manager "${this.name}" failed to install packages`, this.name, { error: err });
        }
        if (installResult.exitCode) {
            this.debug('(install) Failed: %O', installResult);
            throw new error_1.InstallError(`Package manager "${this.name}" failed to install packages`, this.name, { exitCode: installResult.exitCode });
        }
        this.debug('(install) Installed %d packages', packedPkgs.length);
        return installResult;
    }
    async pack(dest, opts = {}) {
        if (!dest) {
            throw new TypeError('(pack) "dest" arg is required');
        }
        let packArgs = [
            'pack',
            '--json',
            `--pack-destination=${dest}`,
            '--foreground-scripts=false', // suppress output of lifecycle scripts so json can be parsed
        ];
        if (opts.workspaces?.length) {
            packArgs = [
                ...packArgs,
                ...opts.workspaces.map((w) => `--workspace=${w}`),
            ];
        }
        else if (opts.allWorkspaces) {
            packArgs = [...packArgs, '--workspaces'];
            if (opts.includeWorkspaceRoot) {
                packArgs = [...packArgs, '--include-workspace-root'];
            }
        }
        let packResult;
        try {
            packResult = await this.executor.exec(packArgs);
        }
        catch (e) {
            const err = e;
            this.debug('(pack) Failed: %O', err);
            // in some cases we can get something more user-friendly via the JSON output
            let parsed;
            try {
                this.debug('Trying to parse stdout: %s', err.stdout);
                parsed = JSON.parse(err.stdout);
            }
            catch {
                // ignore
            }
            if (parsed?.error) {
                throw new error_1.PackError(`Package manager "${this.name}" failed to pack:\n\n${(0, chalk_1.red)(parsed.error.summary)}`, this.name, {
                    error: err,
                    output: err.stderr,
                });
            }
            throw new error_1.PackError(`Package manager "${this.name}" failed to pack`, this.name, {
                error: err,
            });
        }
        // I am not sure why exitCode would be non-zero here.  keep an eye on this;
        // it might never be called
        if (packResult.exitCode) {
            this.debug('(pack) Failed: %O', packResult);
            throw new error_1.PackError(`Package manager "${this.name}" failed to pack`, this.name, {
                exitCode: packResult.exitCode,
                output: packResult.stderr,
            });
        }
        let parsed;
        const { stdout: packOutput } = packResult;
        try {
            parsed = JSON.parse(packOutput);
        }
        catch (err) {
            this.debug('(pack) Failed to parse JSON: %s', packOutput);
            throw new error_1.PackParseError(`Failed to parse JSON result of pack from package manager "${this.name}"`, this.name, err, packOutput);
        }
        const packedPkgs = parsed.map(({ filename, name }) => {
            // workaround for https://github.com/npm/cli/issues/3405
            filename = filename.replace(/^@(.+?)\//, '$1-');
            return {
                tarballFilepath: node_path_1.default.join(dest, filename),
                installPath: node_path_1.default.join(dest, 'node_modules', name),
                pkgName: name,
            };
        });
        this.debug('(pack) Packed %d packages', packedPkgs.length);
        return { packedPkgs, tarballRootDir: dest };
    }
}
exports.Npm7 = Npm7;
Npm7.bin = 'npm';
exports.default = Npm7;
//# sourceMappingURL=npm7.js.map