"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.normalizeVersion = void 0;
const promises_1 = __importDefault(require("node:fs/promises"));
const node_path_1 = __importDefault(require("node:path"));
const semver_1 = require("semver");
const error_1 = require("../error");
const util_1 = require("../util");
async function cachedRead(filename) {
    if (cachedRead.cache.has(filename)) {
        return cachedRead.cache.get(filename);
    }
    const result = JSON.parse(await promises_1.default.readFile(filename, 'utf8'));
    cachedRead.cache.set(filename, result);
    return result;
}
cachedRead.cache = new Map();
async function readDistTags(name) {
    const dataDir = await (0, util_1.findDataDir)();
    const distTagsPath = node_path_1.default.join(dataDir, `${name}-dist-tags.json`);
    let disttags = await cachedRead(distTagsPath);
    if (name === 'yarn') {
        const yarnTags = await cachedRead(node_path_1.default.join(dataDir, 'yarn-tags.json'));
        disttags = { ...disttags, ...yarnTags.latest };
    }
    return disttags;
}
async function readVersions(name) {
    const dataDir = await (0, util_1.findDataDir)();
    const versionsPath = node_path_1.default.join(dataDir, `${name}-versions.json`);
    let versions = JSON.parse(await promises_1.default.readFile(versionsPath, 'utf8'));
    if (name === 'yarn') {
        const yarnTags = await cachedRead(node_path_1.default.join(dataDir, 'yarn-tags.json'));
        versions = [...versions, ...yarnTags.tags];
    }
    return new Set(versions);
}
/**
 * Given a package manager and optionally a version (or dist tag), validate it and return the semver version.
 *
 * This exists because `corepack` expects a SemVer version and not a _range_.  Something like "8" is considered a range. This will find the latest version matching the range.
 * @param name Package manager name
 * @param version Version or dist tag
 * @returns SemVer version string
 */
async function normalizeVersion(name, version) {
    version = version?.length ? version : 'latest';
    const versions = await readVersions(name);
    if ((0, semver_1.valid)(version)) {
        if (versions.has(version)) {
            return (0, semver_1.parse)(version);
        }
        throw new error_1.UnknownVersionError(`Unknown version "${version}" of package manager "${name}"`, name, version);
    }
    const range = (0, semver_1.validRange)(version);
    if (range) {
        const max = (0, semver_1.maxSatisfying)([...versions], range, true);
        if (!max) {
            throw new error_1.UnknownVersionRangeError(`No version found for package manager "${name}" matching range "${version}"`, name, version);
        }
        return (0, semver_1.parse)(max);
    }
    const distTags = await readDistTags(name);
    if (version in distTags) {
        return (0, semver_1.parse)(distTags[version]);
    }
    else {
        throw new error_1.UnknownDistTagError(`Unknown version/tag "${version}" for package manager "${name}"`, name, version);
    }
}
exports.normalizeVersion = normalizeVersion;
//# sourceMappingURL=version.js.map