"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.YarnClassic = void 0;
const debug_1 = __importDefault(require("debug"));
const node_path_1 = __importDefault(require("node:path"));
const error_1 = require("../error");
const util_1 = require("../util");
class YarnClassic {
    constructor(executor, opts = {}) {
        this.executor = executor;
        this.opts = opts;
        this.name = 'yarn';
        this.debug = (0, debug_1.default)(`midnight-smoker:pm:yarn1`);
    }
    static accepts(semver) {
        return Boolean(semver.compare('1.0.0') - semver.compare('2.0.0'));
    }
    static load(executor, opts) {
        return new YarnClassic(executor, opts);
    }
    async install(manifest) {
        const { packedPkgs, tarballRootDir } = manifest;
        if (!packedPkgs?.length) {
            throw new TypeError('(install) Non-empty "packedPkgs" prop in "manifest" arg is required');
        }
        const additionalDeps = manifest.additionalDeps ?? [];
        const installArgs = [
            'add',
            '--no-lockfile',
            '--force',
            ...packedPkgs.map(({ tarballFilepath }) => tarballFilepath),
            ...additionalDeps,
        ];
        let installResult;
        try {
            installResult = await this.executor.exec(installArgs, {
                cwd: tarballRootDir,
            });
        }
        catch (err) {
            throw new error_1.InstallError(`Package manager "${this.name}" failed to install packages`, this.name, { error: err });
        }
        if (installResult.exitCode) {
            this.debug('(install) Failed: %O', installResult);
            throw new error_1.InstallError(`Package manager "${this.name}" failed to install packages`, this.name, { exitCode: installResult.exitCode });
        }
        this.debug('(install) Installed %d packages', packedPkgs.length);
        return installResult;
    }
    async pack(dest, opts = {}) {
        const seenSlugs = new Set();
        const computeSlug = (info) => {
            let slug = node_path_1.default.basename(info.location);
            for (let i = 0; i++; seenSlugs.has(slug)) {
                slug = `${slug}-${i}`;
            }
            seenSlugs.add(slug);
            return slug;
        };
        const finalizePackCommand = (info, pkgName) => {
            const slug = computeSlug(info);
            const tarball = node_path_1.default.join(dest, `${slug}.tgz`);
            const cwd = node_path_1.default.isAbsolute(info.location)
                ? info.location
                : node_path_1.default.join(process.cwd(), info.location);
            return {
                command: [...basePackArgs, `--filename=${tarball}`],
                cwd,
                tarball,
                pkgName,
            };
        };
        const getWorkspaceRootPackageName = async () => {
            const { packageJson } = await (0, util_1.readPackageJson)({
                cwd: process.cwd(),
                strict: true,
            });
            const { name = node_path_1.default.dirname(process.cwd()) } = packageJson;
            return name;
        };
        if (!dest) {
            throw new TypeError('(pack) "dest" arg is required');
        }
        const commands = [];
        const basePackArgs = ['pack', '--json'];
        const shouldUseWorkspaces = Boolean(opts.allWorkspaces || opts.workspaces?.length);
        if (shouldUseWorkspaces) {
            let workspaceInfo;
            try {
                let { stdout } = await this.executor.exec(['workspaces', 'info']);
                const lines = stdout.split(/\r?\n/);
                lines.shift();
                lines.pop();
                stdout = lines.join('\n');
                workspaceInfo = JSON.parse(stdout);
            }
            catch (err) {
                throw new error_1.PackError(`Package manager "${this.name}" unable to read workspace information`, this.name, { error: err });
            }
            if (opts.workspaces?.length) {
                commands.push(...opts.workspaces.map((workspace) => {
                    let info = workspaceInfo[workspace];
                    let name;
                    if (!info) {
                        [name, info] = Object.entries(workspaceInfo).find(([, info]) => info.location === workspace);
                        if (!info) {
                            throw new error_1.PackError(`Package manager "${this.name}" unable to find workspace "${workspace}"`, this.name);
                        }
                    }
                    else {
                        name = workspace;
                    }
                    return finalizePackCommand(info, name);
                }));
            }
            else {
                // allWorkspaces must be true
                commands.push(...Object.entries(workspaceInfo).map(([name, info]) => finalizePackCommand(info, name)));
                if (opts.includeWorkspaceRoot) {
                    commands.push(finalizePackCommand({ location: process.cwd() }, await getWorkspaceRootPackageName()));
                }
            }
        }
        else {
            commands.push(finalizePackCommand({ location: process.cwd() }, await getWorkspaceRootPackageName()));
        }
        this.debug(commands);
        const packedPkgs = [];
        for await (const { command, cwd, tarball, pkgName } of commands) {
            let packResult;
            try {
                packResult = await this.executor.exec(command, { cwd });
            }
            catch (err) {
                throw new error_1.PackError(`Package manager "${this.name}" failed to pack`, this.name, {
                    error: err,
                });
            }
            if (packResult.exitCode) {
                this.debug('(pack) Failed: %O', packResult);
                throw new error_1.PackError(`Package manager "${this.name}" failed to pack`, this.name, {
                    exitCode: packResult.exitCode,
                    output: packResult.stderr,
                });
            }
            packedPkgs.push({
                tarballFilepath: tarball,
                installPath: node_path_1.default.join(dest, 'node_modules', pkgName),
                pkgName,
            });
        }
        this.debug('(pack) Packed %d packages', packedPkgs.length);
        return { packedPkgs, tarballRootDir: dest };
    }
    async runScript(manifest) {
        if (!manifest) {
            throw new TypeError('(runScript) "manifest" arg is required');
        }
        const { script, packedPkg } = manifest;
        const args = ['run', script];
        const { pkgName, installPath: cwd } = packedPkg;
        let result;
        try {
            const rawResult = await this.executor.exec(args, {
                cwd,
            });
            result = { pkgName, script, rawResult, cwd };
        }
        catch (err) {
            const error = err;
            result = {
                pkgName,
                script,
                rawResult: error,
                cwd,
            };
            if (this.opts.loose && /Command ".+?" not found/i.test(error.stderr)) {
                result.skipped = true;
            }
            else {
                result.error = new error_1.RunScriptError(`Script "${script}" in package "${pkgName}" failed`, script, pkgName, this.name, { error, exitCode: error.exitCode, output: error.stderr });
            }
        }
        if (!result.error && !result.skipped && result.rawResult.failed) {
            let message;
            if (result.rawResult.stderr &&
                /Command ".+?" not found/i.test(result.rawResult.stderr)) {
                message = `(runScript) Script "${script}" in package "${pkgName}" failed; script not found`;
            }
            else {
                if (result.rawResult.exitCode) {
                    message = `(runScript) Script "${script}" in package "${pkgName}" failed with exit code ${result.rawResult.exitCode}: ${result.rawResult.all}`;
                }
                else {
                    message = `(runScript) Script "${script}" in package "${pkgName}" failed: ${result.rawResult.all}`;
                }
            }
            result.error = new error_1.RunScriptError(message, script, pkgName, this.name, {
                exitCode: result.rawResult.exitCode,
                output: result.rawResult.all,
            });
        }
        if (result.error) {
            this.debug(`(runScripts) Script "%s" in package "%s" failed; continuing...`, script, pkgName);
        }
        else {
            this.debug('(runScripts) Successfully executed script %s in package %s', script, pkgName);
        }
        return result;
    }
}
exports.YarnClassic = YarnClassic;
YarnClassic.bin = 'yarn';
exports.default = YarnClassic;
//# sourceMappingURL=yarn-classic.js.map