"use strict";
/**
 * `no-banned-files` checks for banned files in the package.
 *
 * Portions of this adapted from [ban-sensitive-files](https://github.com/bahmutov/ban-sensitive-files), including `git-deny-patterns.json` and {@linkcode reToRegExp}.
 * @module
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const promises_1 = __importDefault(require("node:fs/promises"));
const node_path_1 = __importDefault(require("node:path"));
const zod_1 = require("zod");
const schema_util_1 = require("../../schema-util");
const util_1 = require("../../util");
const rule_1 = require("../rule");
const debug = (0, debug_1.default)('midnight-smoker:rule:no-banned-files');
let denyPatterns;
/**
 * Load and cache the deny patterns
 */
async function loadDenyPatterns() {
    if (denyPatterns) {
        return denyPatterns;
    }
    const dataDir = await (0, util_1.findDataDir)();
    const patterns = await promises_1.default.readFile(node_path_1.default.join(dataDir, 'git-deny-patterns.json'), 'utf-8');
    denyPatterns = JSON.parse(patterns);
    return denyPatterns;
}
const regExMap = new Map();
/**
 * Just caching some `RegExp` objects.
 * @param pattern RegExp as a string
 * @returns A RegExp
 */
function getRegExp(pattern) {
    if (!regExMap.has(pattern)) {
        regExMap.set(pattern, new RegExp(pattern));
    }
    return regExMap.get(pattern);
}
/**
 * Uses a list of {@linkcode DenyPattern} objects to check if a filename is banned
 * @param filename Filename to check
 * @returns If the file is banned, a string with a description; otherwise, `undefined`
 */
async function isBannedFilename(filename) {
    denyPatterns = await loadDenyPatterns();
    for (const { part, type, pattern, caption } of denyPatterns) {
        const value = part === 'filename'
            ? filename
            : node_path_1.default.extname(filename).replace(/^\./, '');
        if ((type === 'match' && value === pattern) ||
            (type === 'regex' && getRegExp(pattern).test(value))) {
            return caption;
        }
    }
}
const noBannedFiles = (0, rule_1.createRule)({
    async check({ pkgPath, fail }, opts) {
        const queue = [pkgPath];
        const failed = [];
        const allow = new Set(opts.allow);
        const deny = new Set(opts.deny);
        while (queue.length) {
            const dir = queue.pop();
            debug('Reading directory %s', dir);
            const dirents = await promises_1.default.readdir(dir, { withFileTypes: true });
            for (const dirent of dirents) {
                if (dirent.isDirectory()) {
                    if (node_path_1.default.basename(dirent.name) === 'node_modules') {
                        continue;
                    }
                    queue.push(node_path_1.default.join(dir, dirent.name));
                }
                else {
                    if (allow.has(dirent.name)) {
                        continue;
                    }
                    if (deny.has(dirent.name)) {
                        failed.push(fail(`Banned file found: ${dirent.name} (per custom deny list)`));
                    }
                    else {
                        const banned = await isBannedFilename(dirent.name);
                        if (banned) {
                            failed.push(fail(`Banned file found: ${dirent.name} (${banned})`));
                        }
                    }
                }
            }
        }
        return failed;
    },
    name: 'no-banned-files',
    description: 'Bans certain files from being published',
    schema: zod_1.z.object({
        allow: schema_util_1.zNonEmptyStringArray.describe('Allow these banned files'),
        deny: schema_util_1.zNonEmptyStringArray.describe('Deny these additional files'),
    }),
});
exports.default = noBannedFiles;
//# sourceMappingURL=no-banned-files.js.map