"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const promises_1 = __importDefault(require("node:fs/promises"));
const node_path_1 = __importDefault(require("node:path"));
const rule_1 = require("../rule");
const zod_1 = require("zod");
const DEFAULT_FIELD = 'main';
/**
 * Default entry points for CJS packages which do not contain a `main` field.
 *
 * Order matters.
 * @see {@link https://nodejs.org/api/modules.html#all-together}
 */
const DEFAULT_ENTRY_POINTS = ['index.js', 'index.json', 'index.node'];
const noMissingEntryPoint = (0, rule_1.createRule)({
    async check({ pkgJson, pkgPath, fail }) {
        // skip if the package has an "exports" field or it's an ESM package
        if ('exports' in pkgJson ||
            ('type' in pkgJson && pkgJson.type === 'module')) {
            return;
        }
        // I don't think this should ever be anything other than `main`; change later if needed
        const field = DEFAULT_FIELD;
        if (pkgJson[field]) {
            // the field exists, so check the file exists
            const relativePath = pkgJson[field];
            const filepath = node_path_1.default.resolve(pkgPath, relativePath);
            try {
                await promises_1.default.stat(filepath);
            }
            catch {
                return [
                    fail(`No entry point found for package "${pkgJson.name}"; file from field "${field}" unreadable at path: ${relativePath}`),
                ];
            }
        }
        else {
            // the field doesn't exist, so look at the default entry points in order of precedence
            let found = false;
            const queue = [...DEFAULT_ENTRY_POINTS];
            while (queue.length && !found) {
                const relativePath = queue.shift(); // in order!!!
                const filepath = node_path_1.default.resolve(pkgPath, relativePath);
                try {
                    await promises_1.default.stat(filepath);
                    found = true;
                }
                catch { }
            }
            if (!found) {
                return [
                    fail(`No entry point found for package "${pkgJson.name}"; (index.js, index.json or index.node)`),
                ];
            }
        }
    },
    schema: zod_1.z.any(),
    name: 'no-missing-entry-point',
    description: 'Checks that the package contains an entry point; only applies to CJS packages without an "exports" field',
});
exports.default = noMissingEntryPoint;
//# sourceMappingURL=no-missing-entry-point.js.map