"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const promises_1 = __importDefault(require("node:fs/promises"));
const node_path_1 = __importDefault(require("node:path"));
const zod_1 = require("zod");
const rule_1 = require("../rule");
const schema_util_1 = require("../../schema-util");
const debug = (0, debug_1.default)('midnight-smoker:rule:no-missing-pkg-files');
/**
 * Each of these is a flag in the options object, which can be set to `false` to
 * avoid checking the value (relative filepath) exists.
 */
const EXPLICIT_FIELD_FLAGS = [
    'bin',
    'browser',
    'types',
    'unpkg',
    'module',
];
const noMissingPkgFiles = (0, rule_1.createRule)({
    async check({ pkgJson, pkgPath, fail }, opts) {
        let fields = opts.fields ?? [];
        // add any explicit fields to the list of fields to check.
        // these are kept separate so someone doesn't overwrite them by providing `fields`
        for (const field of EXPLICIT_FIELD_FLAGS) {
            if (opts[field] !== false) {
                fields.push(field);
            }
        }
        fields = [...new Set(fields)]; // unique
        debug(`Checking fields: %O`, fields);
        /**
         * @param relativePath Path from package root to file
         * @param field Field name
         * @param name Optional; refers to key in an object field (e.g., `bin` as an object instead of `string`)
         * @todo maybe make this a utility function?
         */
        const checkFile = async (relativePath, field, name) => {
            const filepath = node_path_1.default.resolve(pkgPath, relativePath);
            debug('Checking for %s from field %s', filepath, field);
            try {
                await promises_1.default.stat(filepath);
            }
            catch {
                return fail(name
                    ? `File "${name}" from "${field}" field unreadable at path: ${relativePath}`
                    : `File from "${field}" field unreadable at path: ${relativePath}`);
            }
        };
        /**
         * Assuming a relative filepath is the value of a field in `package.json` _or_ the value
         * is a `Record<string, string>` (e.g., `bin`), return an array of {@linkcode PkgFilepathInfo} objects to check for existence.
         * @param currentInfo - List of {@linkcode PkgFilepathInfo} objects to append to
         * @param field - Fieldname
         * @returns New array with any new objects appended
         */
        const filenameReducer = (currentInfo, field) => {
            let newInfo = [...currentInfo];
            if (field in pkgJson) {
                const relativePath = pkgJson[field];
                if (relativePath) {
                    if (typeof relativePath === 'string') {
                        newInfo = [...newInfo, { relativePath, field }];
                    }
                    else if (!Array.isArray(relativePath) &&
                        typeof relativePath === 'object') {
                        newInfo = [
                            ...newInfo,
                            ...Object.entries(relativePath)
                                .filter(([, relativePath]) => typeof relativePath === 'string')
                                .map(([name, relativePath]) => ({
                                relativePath: relativePath,
                                field,
                                name,
                            })),
                        ];
                    }
                }
            }
            return newInfo;
        };
        /**
         * Array of {@linkcode CheckFailure} or `undefined` (if the file exists)
         */
        const res = await Promise.all(fields
            .reduce(filenameReducer, [])
            .map(({ relativePath, field, name }) => checkFile(relativePath, field, name)));
        return res.filter(Boolean);
    },
    name: 'no-missing-pkg-files',
    description: 'Checks that files referenced in package.json exist in the tarball',
    schema: zod_1.z.object({
        bin: schema_util_1.zTrue.describe('Check the "bin" field (if it exists)'),
        browser: schema_util_1.zTrue.describe('Check the "browser" field (if it exists)'),
        types: schema_util_1.zTrue.describe('Check the "types" field (if it exists)'),
        unpkg: schema_util_1.zTrue.describe('Check the "unpkg" field (if it exists)'),
        module: schema_util_1.zTrue.describe('Check the "module" field (if it exists)'),
        fields: schema_util_1.zNonEmptyStringArray.describe('Check files referenced by these additional top-level fields'),
    }),
});
exports.default = noMissingPkgFiles;
//# sourceMappingURL=no-missing-pkg-files.js.map