import type { PackageJson } from 'read-pkg-up';
import { z } from 'zod';
import type { CheckFailure } from './result';
import { type CheckSeverity } from './severity';
/**
 * The bits of a {@linkcode CheckContext} suitable for serialization.
 * @internal
 */
export interface StaticCheckContext {
    pkgJson: PackageJson;
    pkgJsonPath: string;
    pkgPath: string;
    severity: CheckSeverity;
}
/**
 * The `fail` function that a {@linkcode RuleCheckFn} uses to create a
 * {@linkcode CheckFailure}. The {@linkcode RuleCheckFn} then returns an array
 * of these.
 */
export type CheckContextFailFn = (message: string, data?: any) => CheckFailure;
/**
 * A context object which is provided to a {@linkcode RuleCheckFn}, containing
 * information about the current package to be checked and how to report a
 * failure.
 *
 * @public
 */
export declare class CheckContext<const Name extends string = string, Schema extends z.ZodTypeAny = z.ZodTypeAny> implements StaticCheckContext {
    /**
     * The (parse as )d `package.json` for the package being checked.
     *
     * _Not_ normalized.
     */
    readonly pkgJson: PackageJson;
    /**
     * The absolute path to this context's package's `package.json`.
     */
    readonly pkgJsonPath: string;
    /**
     * The absolute path to this context's package's root directory.
     *
     * This is the same as `path.dirname(pkgJsonPath)`
     */
    readonly pkgPath: string;
    /**
     * The severity level for this rule (as chosen by the user)
     */
    readonly severity: CheckSeverity;
    /**
     * The `fail` function as provided to the `Rule`'s {@linkcode RuleCheckFn}.
     */
    readonly fail: CheckContextFailFn;
    constructor(rule: Rule<Name, Schema>, staticCtx: StaticCheckContext);
    /**
     * Creates a `fail` function for {@linkcode RuleCheckFn}.
     * @param rule The `Rule` which will receive the `fail` function
     * @returns A `fail` function
     */
    private createFailFn;
    /**
     * Omits the {@linkcode CheckContext.pkgJson} property (too big).
     *
     * @returns A JSON-serializable representation of this object
     */
    toJSON(): {
        pkgJsonPath: string;
        pkgPath: string;
        severity: "error" | "warn" | "off";
    };
}
/**
 * The bits of a {@linkcode RuleDef} suitable for passing the API edge
 */
export interface StaticRuleDef<Name extends string = string> {
    readonly defaultSeverity?: CheckSeverity;
    readonly description: string;
    readonly name: Name;
}
/**
 * The type _returned or fulfilled_ by the {@linkcode Rule.check} method.
 */
export type RuleCheckFnResult = CheckFailure[] | undefined;
export type RuleOptions<Schema extends z.ZodTypeAny> = z.infer<Schema>;
/**
 * The function which actually performs the check within a {@linkcode Rule}
 * @public
 */
export type RuleCheckFn<Name extends string, Schema extends z.ZodTypeAny> = (ctx: CheckContext<Name, Schema>, opts: RuleOptions<Schema>) => Promise<RuleCheckFnResult> | RuleCheckFnResult;
/**
 * The raw definition of a {@linkcode Rule}, as defined by a implementor.
 * @public
 */
export interface RuleDef<Name extends string, Schema extends z.ZodTypeAny> extends StaticRuleDef<Name> {
    /**
     * The function which actually performs the check.
     */
    check: RuleCheckFn<Name, Schema>;
    /**
     * Options schema for this rule, if any
     */
    schema: Schema;
}
/**
 * Represents a _Rule_, which performs a check upon an installed (from tarball) package.
 * @internal
 */
export declare class Rule<const Name extends string, Schema extends z.ZodTypeAny> implements RuleDef<Name, Schema> {
    /**
     * The name for this rule.
     *
     * @todo enforce uniqueness
     */
    readonly name: Name;
    /**
     * The description for this rule
     */
    readonly description: string;
    /**
     * The default severity for this rule if not supplied by the user
     */
    readonly defaultSeverity: CheckSeverity;
    /**
     * The function which actually performs the check.
     */
    readonly check: RuleCheckFn<Name, Schema>;
    /**
     * The options schema for this rule, if any
     */
    readonly schema: Schema;
    /**
     * A composable schema handling the default severity for this rule
     */
    private readonly zDefaultRuleSeverity;
    readonly defaultOptions: RuleOptions<Schema>;
    constructor(def: RuleDef<Name, Schema>);
    /**
     * Returns the entire schema for the value of this rule in the `RuleConfig` object.
     */
    get zRuleSchema(): z.ZodUnion<[z.ZodEffects<z.ZodDefault<z.ZodDefault<z.ZodEnum<["off", "warn", "error"]>>>, {
        severity: "error" | "warn" | "off";
        opts: RuleOptions<Schema>;
    }, "error" | "warn" | "off" | undefined>, z.ZodEffects<Schema, {
        severity: "error" | "warn" | "off";
        opts: any;
    }, z.input<Schema>>, z.ZodEffects<z.ZodTuple<[Schema, z.ZodDefault<z.ZodDefault<z.ZodEnum<["off", "warn", "error"]>>>], null>, {
        severity: "error" | "warn" | "off";
        opts: Schema extends z.ZodType<any, any, any> ? Schema["_output"] : never;
    }, [Schema extends z.ZodType<any, any, any> ? Schema["_input"] : never, "error" | "warn" | "off" | undefined]>]>;
    /**
     * Creates a {@linkcode RuleCont}.
     *
     * Wrap a {@linkcode Rule} in this function to allow it to be used in a
     * collection of some sort.
     */
    toRuleCont(): <R>(cont: <const Name_1 extends string, Schema_1 extends z.ZodTypeAny = z.ZodTypeAny>(rule: Rule<Name_1, Schema_1>) => R) => R;
    /**
     * Returns this `Rule` in a format suitable for serialization.
     */
    toJSON(): {
        name: Name;
        description: string;
    };
}
/**
 * A continuation for a {@link Rule}.
 *
 * This devilry is an for an existential type; it allows us to iterate over a
 * list of rules having different generic type parameters.
 *
 * @internal
 * @see {@link https://jalo.website/existential-types-in-typescript-through-continuations}
 */
export type RuleCont = <R>(cont: <const Name extends string, Schema extends z.ZodTypeAny = z.ZodTypeAny>(rule: Rule<Name, Schema>) => R) => R;
/**
 * {@linkcode Rule} factory.
 *
 * Rules use this to create themselves from raw {@linkcode RuleDef RuleDefs}.
 * @param ruleDef - Raw rule definition
 * @returns A new {@linkcode Rule}
 */
export declare function createRule<const Name extends string, Schema extends z.ZodTypeAny = z.ZodTypeAny>(ruleDef: RuleDef<Name, Schema>): Rule<Name, Schema>;
//# sourceMappingURL=rule.d.ts.map