"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createRule = exports.Rule = exports.CheckContext = void 0;
const zod_1 = require("zod");
const severity_1 = require("./severity");
/**
 * A context object which is provided to a {@linkcode RuleCheckFn}, containing
 * information about the current package to be checked and how to report a
 * failure.
 *
 * @public
 */
class CheckContext {
    constructor(rule, staticCtx) {
        this.pkgJson = staticCtx.pkgJson;
        this.pkgJsonPath = staticCtx.pkgJsonPath;
        this.pkgPath = staticCtx.pkgPath;
        this.severity = staticCtx.severity;
        this.fail = this.createFailFn(rule);
    }
    /**
     * Creates a `fail` function for {@linkcode RuleCheckFn}.
     * @param rule The `Rule` which will receive the `fail` function
     * @returns A `fail` function
     */
    createFailFn(rule) {
        return (message, data) => ({
            rule: rule.toJSON(),
            message,
            data,
            context: this,
            failed: true,
            severity: this.severity,
        });
    }
    /**
     * Omits the {@linkcode CheckContext.pkgJson} property (too big).
     *
     * @returns A JSON-serializable representation of this object
     */
    toJSON() {
        return {
            pkgJsonPath: this.pkgJsonPath,
            pkgPath: this.pkgPath,
            severity: this.severity,
        };
    }
}
exports.CheckContext = CheckContext;
/**
 * Represents a _Rule_, which performs a check upon an installed (from tarball) package.
 * @internal
 */
class Rule {
    constructor(def) {
        this.name = def.name;
        this.description = def.description;
        this.defaultSeverity = def.defaultSeverity ?? severity_1.CheckSeverities.ERROR;
        this.check = def.check;
        this.schema = def.schema;
        this.zDefaultRuleSeverity = severity_1.zCheckSeverity.default(this.defaultSeverity);
        this.defaultOptions = this.schema.parse({});
    }
    /**
     * Returns the entire schema for the value of this rule in the `RuleConfig` object.
     */
    get zRuleSchema() {
        const { zDefaultRuleSeverity, schema: zSchema } = this;
        return zod_1.z.union([
            zDefaultRuleSeverity.transform((severity) => ({
                severity,
                opts: this.defaultOptions,
            })),
            zSchema.transform((opts) => ({ severity: this.defaultSeverity, opts })),
            zod_1.z
                .tuple([zSchema, zDefaultRuleSeverity])
                .transform(([opts, severity]) => ({ severity, opts })),
        ]);
    }
    /**
     * Creates a {@linkcode RuleCont}.
     *
     * Wrap a {@linkcode Rule} in this function to allow it to be used in a
     * collection of some sort.
     */
    toRuleCont() {
        return (cont) => cont(this);
    }
    /**
     * Returns this `Rule` in a format suitable for serialization.
     */
    toJSON() {
        return {
            name: this.name,
            description: this.description,
        };
    }
}
exports.Rule = Rule;
/**
 * {@linkcode Rule} factory.
 *
 * Rules use this to create themselves from raw {@linkcode RuleDef RuleDefs}.
 * @param ruleDef - Raw rule definition
 * @returns A new {@linkcode Rule}
 */
function createRule(ruleDef) {
    return new Rule(ruleDef);
}
exports.createRule = createRule;
//# sourceMappingURL=rule.js.map