/// <reference types="node" />
import { EventEmitter } from 'node:events';
import StrictEventEmitter from 'strict-event-emitter-types';
import { type SmokerEvent } from './event';
import { type RawSmokerOptions, type SmokerOptions } from './options';
import { type InstallResults, type PackageManager } from './pm';
import { type CheckOptions, type RuleCont } from './rules';
import { type CheckResults } from './rules/result';
import { type PkgInstallManifest, type PkgRunManifest, type RunScriptResult, type SmokeResults } from './types';
export declare const TMP_DIR_PREFIX = "midnight-smoker-";
type TSmokerEmitter = StrictEventEmitter<EventEmitter, SmokerEvent>;
declare const Smoker_base: new () => TSmokerEmitter;
export declare class Smoker extends Smoker_base {
    readonly pms: Map<string, PackageManager>;
    /**
     * List of extra dependencies to install
     */
    private readonly add;
    /**
     * Whether to run against all workspaces
     */
    private readonly allWorkspaces;
    /**
     * Whether to bail on the first script failure
     */
    private readonly bail;
    /**
     * Whether to include the workspace root
     */
    private readonly includeWorkspaceRoot;
    /**
     * Whether to keep temp dirs around (debugging purposes)
     */
    private readonly linger;
    /**
     * Mapping of {@linkcode PackageManager} instances to their identifiers of the form `<npm|yarn|pnpm>@<version>`
     */
    private readonly pmIds;
    /**
     * List of temp directories created
     */
    private readonly tempDirs;
    /**
     * List of specific workspaces to run against
     */
    private readonly workspaces;
    readonly ruleConfig: CheckOptions;
    /**
     * List of scripts to run in each workspace
     */
    readonly scripts: string[];
    /**
     * Whether or not to run builtin checks
     */
    private readonly checks;
    private readonly originalOpts;
    private constructor();
    static init(opts?: RawSmokerOptions): Promise<Smoker>;
    static create(pms: Map<string, PackageManager>, opts?: RawSmokerOptions | SmokerOptions): Smoker;
    /**
     * Run the smoke test scripts!
     * @param opts - Options
     */
    static smoke(opts?: RawSmokerOptions): Promise<SmokeResults>;
    /**
     * Cleans up any temp directories created by {@linkcode createTempDir}.
     *
     * If the {@linkcode SmokeOptions.linger} option is set to `true`, this method
     * will _not_ clean up the directories, but will instead emit a
     * {@linkcode SmokerEvent.Lingered|Lingered} event.
     */
    cleanup(): Promise<void>;
    /**
     * Creates a temp dir and adds it to the set of temp dirs to be cleaned up later
     * @returns New temp dir path
     */
    createTempDir(): Promise<string>;
    /**
     * Installs from tarball in a temp dir
     */
    install(manifests: PkgInstallManifest): Promise<InstallResults>;
    /**
     * For each package manager, creates a tarball for one or more packages
     */
    pack(): Promise<PkgInstallManifest>;
    /**
     * @internal
     * @param ruleCont - Rule continuation
     * @param pkgPath - Path to installed package
     * @param checkOpts - Parsed rule options
     * @returns Results of a single check
     */
    runCheck(ruleCont: RuleCont, pkgPath: string, checkOpts: CheckOptions): Promise<CheckResults>;
    runChecks(installResults: InstallResults): Promise<CheckResults>;
    /**
     * Runs the script for each package in `packItems`
     */
    runScripts(pkgRunManifest: PkgRunManifest): Promise<RunScriptResult[]>;
    /**
     * Converts install results to a {@linkcode PkgRunManifest} for {@linkcode runScripts}
     * @param installResults Results of {@linkcode install}
     * @returns Package managers and scripts to run for each unpacked package
     */
    private buildPkgRunManifest;
    /**
     * Returns `true` if any of the scripts failed or any of the checks failed
     * @param results Results from {@linkcode smoke}
     * @returns Whether the results indicate a failure
     */
    isSmokeFailure({ scripts, checks }: SmokeResults): boolean;
    /**
     * Pack, install, run checks (optionally), and run scripts (optionally)
     * @returns Results
     */
    smoke(): Promise<SmokeResults>;
    /**
     * This is only here because it's a fair amount of work to mash the data into a format more suitable for display.
     *
     * This is used by the events {@linkcode SmokerEvent.InstallBegin}, {@linkcode SmokerEvent.InstallOk}, and {@linkcode SmokerEvent.PackOk}.
     * @param pkgInstallManifest What to install and with what package manager
     * @returns Something to be emitted
     */
    private buildInstallEventData;
}
export {};
//# sourceMappingURL=smoker.d.ts.map