"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Smoker = exports.TMP_DIR_PREFIX = void 0;
/* eslint-disable no-labels */
const debug_1 = __importDefault(require("debug"));
const node_events_1 = require("node:events");
const promises_1 = __importDefault(require("node:fs/promises"));
const node_os_1 = require("node:os");
const node_path_1 = __importDefault(require("node:path"));
const error_1 = require("./error");
const event_1 = require("./event");
const options_1 = require("./options");
const pm_1 = require("./pm");
const rules_1 = require("./rules");
const builtin_1 = require("./rules/builtin");
const util_1 = require("./util");
const debug = (0, debug_1.default)('midnight-smoker:smoker');
exports.TMP_DIR_PREFIX = 'midnight-smoker-';
function createStrictEventEmitterClass() {
    const TypedEmitter = node_events_1.EventEmitter;
    return TypedEmitter;
}
class Smoker extends createStrictEventEmitterClass() {
    constructor(pms, opts) {
        super();
        this.pms = pms;
        const { script, linger, includeRoot, add, bail, all, workspace, checks, rules, } = opts;
        this.originalOpts = Object.freeze(opts);
        this.scripts = script;
        this.linger = linger;
        this.includeWorkspaceRoot = includeRoot;
        this.add = add;
        this.bail = bail;
        this.allWorkspaces = all;
        this.workspaces = workspace;
        this.checks = checks;
        this.ruleConfig = rules;
        if (this.allWorkspaces && this.workspaces.length) {
            throw new error_1.InvalidArgError('Option "workspace" is mutually exclusive with "all" and/or "includeRoot"');
        }
        this.pmIds = new WeakMap();
        for (const [pmId, pm] of pms) {
            this.pmIds.set(pm, pmId);
        }
        this.tempDirs = new Set();
    }
    static async init(opts = {}) {
        const smokerOpts = (0, options_1.parseOptions)(opts);
        const { verbose, pm, loose } = smokerOpts;
        const pms = await (0, pm_1.loadPackageManagers)(pm, {
            verbose,
            loose,
        });
        return Smoker.create(pms, smokerOpts);
    }
    static create(pms, opts = {}) {
        return new Smoker(pms, (0, options_1.parseOptions)(opts));
    }
    /**
     * Run the smoke test scripts!
     * @param opts - Options
     */
    static async smoke(opts = {}) {
        const smoker = await Smoker.init(opts);
        return smoker.smoke();
    }
    /**
     * Cleans up any temp directories created by {@linkcode createTempDir}.
     *
     * If the {@linkcode SmokeOptions.linger} option is set to `true`, this method
     * will _not_ clean up the directories, but will instead emit a
     * {@linkcode SmokerEvent.Lingered|Lingered} event.
     */
    async cleanup() {
        if (!this.linger) {
            await Promise.all([...this.tempDirs].map(async (tempdir) => {
                try {
                    await promises_1.default.rm(tempdir, { recursive: true, force: true });
                }
                catch (e) {
                    const err = e;
                    if (err.code !== 'ENOENT') {
                        throw new error_1.DirDeletionError(`Failed to clean temp directory ${tempdir}`, tempdir, err);
                    }
                }
                finally {
                    this.tempDirs.delete(tempdir);
                }
            }));
        }
        else if (this.tempDirs.size) {
            await Promise.resolve();
            this.emit(event_1.Event.LINGERED, [...this.tempDirs]);
        }
    }
    /**
     * Creates a temp dir and adds it to the set of temp dirs to be cleaned up later
     * @returns New temp dir path
     */
    async createTempDir() {
        // TODO EMIT
        try {
            const prefix = node_path_1.default.join((0, node_os_1.tmpdir)(), exports.TMP_DIR_PREFIX);
            const tempdir = await promises_1.default.mkdtemp(prefix);
            this.tempDirs.add(tempdir);
            return tempdir;
        }
        catch (err) {
            throw new error_1.DirCreationError('Failed to create temp directory', err);
        }
    }
    /**
     * Installs from tarball in a temp dir
     */
    async install(manifests) {
        if (!manifests?.size) {
            throw new error_1.InvalidArgError('Non-empty "manifests" arg is required', 'manifests');
        }
        // ensure we emit asynchronously
        await Promise.resolve();
        const installData = this.buildInstallEventData(manifests);
        this.emit(event_1.Event.INSTALL_BEGIN, installData);
        const installResults = new Map();
        for (const [pm, manifest] of manifests) {
            try {
                const manifestWithAdds = {
                    ...manifest,
                    additionalDeps: this.add ?? [],
                };
                debug('Installing package(s) %O using %s', [
                    ...manifestWithAdds.packedPkgs.map((p) => p.pkgName),
                    ...manifestWithAdds.additionalDeps,
                ], this.pmIds.get(pm));
                const result = await pm.install(manifestWithAdds);
                installResults.set(pm, [manifest, result]);
            }
            catch (err) {
                const error = err;
                this.emit(event_1.Event.INSTALL_FAILED, error);
                throw error;
            }
        }
        this.emit(event_1.Event.INSTALL_OK, installData);
        return installResults;
    }
    /**
     * For each package manager, creates a tarball for one or more packages
     */
    async pack() {
        await Promise.resolve();
        this.emit(event_1.Event.PACK_BEGIN, { packageManagers: [...this.pms.keys()] });
        const manifestMap = new Map();
        for await (const pm of this.pms.values()) {
            const dest = await this.createTempDir();
            let manifest;
            try {
                debug('Packing into %s using %s', dest, this.pmIds.get(pm));
                manifest = await pm.pack(dest, {
                    allWorkspaces: this.allWorkspaces,
                    includeWorkspaceRoot: this.includeWorkspaceRoot,
                    workspaces: this.workspaces,
                });
                manifestMap.set(pm, manifest);
            }
            catch (err) {
                this.emit(event_1.Event.PACK_FAILED, err);
                throw err;
            }
        }
        this.emit(event_1.Event.PACK_OK, this.buildInstallEventData(manifestMap));
        return manifestMap;
    }
    /**
     * @internal
     * @param ruleCont - Rule continuation
     * @param pkgPath - Path to installed package
     * @param checkOpts - Parsed rule options
     * @returns Results of a single check
     */
    async runCheck(ruleCont, pkgPath, checkOpts) {
        return ruleCont(async (rule) => {
            const { name: ruleName } = rule;
            const ruleOpts = checkOpts[ruleName];
            const { severity, opts } = ruleOpts;
            // XXX might be something better to do here. this won't happen during
            // expected non-test operation
            /* istanbul ignore next */
            if (severity === rules_1.CheckSeverities.OFF) {
                return { failed: [], passed: [] };
            }
            const { packageJson: pkgJson, path: pkgJsonPath } = await (0, util_1.readPackageJson)({
                cwd: pkgPath,
                strict: true,
            });
            const staticCtx = {
                pkgJson,
                pkgJsonPath,
                pkgPath,
                severity,
            };
            debug(`Running rule %s with context %O and opts %O`, ruleName, {
                pkgJsonPath,
                pkgPath,
                severity,
                pkgName: pkgJson.name,
            }, opts);
            const context = new rules_1.CheckContext(rule, staticCtx);
            let result;
            try {
                result = await rule.check(context, opts);
            }
            catch (err) {
                this.emit(event_1.Event.RULE_ERROR, new error_1.RuleError(`Rule "${ruleName}" threw an exception`, staticCtx, ruleName, err));
                return {
                    failed: [
                        {
                            message: String(err),
                            failed: true,
                            severity,
                            rule,
                            context: staticCtx,
                        },
                    ],
                    passed: [],
                };
            }
            if (result?.length) {
                return {
                    failed: result,
                    passed: [],
                };
            }
            const ok = {
                rule,
                context,
                failed: false,
            };
            return { failed: [], passed: [ok] };
        });
    }
    async runChecks(installResults) {
        const allFailed = [];
        const allPassed = [];
        const { ruleConfig } = this;
        await Promise.resolve();
        const runnableRules = builtin_1.BuiltinRuleConts.filter((ruleCont) => ruleCont((rule) => ruleConfig[rule.name].severity !==
            rules_1.CheckSeverities.OFF));
        const total = runnableRules.length;
        let current = 0;
        this.emit(event_1.Event.RUN_CHECKS_BEGIN, { config: ruleConfig, total });
        // run against multiple package managers??
        for (const ruleCont of runnableRules) {
            const ruleName = ruleCont((rule) => rule.name);
            const configForRule = ruleConfig[ruleName];
            this.emit(event_1.Event.RUN_CHECK_BEGIN, {
                rule: ruleName,
                config: configForRule,
                current,
                total,
            });
            for (const [{ packedPkgs }] of installResults.values()) {
                for (const { installPath: pkgPath } of packedPkgs) {
                    const { failed, passed } = await this.runCheck(ruleCont, pkgPath, ruleConfig);
                    if (failed.length) {
                        this.emit(event_1.Event.RUN_CHECK_FAILED, {
                            rule: ruleName,
                            config: configForRule,
                            current,
                            total,
                            failed,
                        });
                    }
                    else {
                        this.emit(event_1.Event.RUN_CHECK_OK, {
                            rule: ruleName,
                            config: configForRule,
                            current,
                            total,
                        });
                    }
                    allFailed.push(...failed);
                    allPassed.push(...passed);
                    current++;
                }
            }
        }
        const evtData = {
            config: ruleConfig,
            total,
            passed: allPassed,
            failed: allFailed,
        };
        if (allFailed.length) {
            this.emit(event_1.Event.RUN_CHECKS_FAILED, evtData);
        }
        else {
            this.emit(event_1.Event.RUN_CHECKS_OK, evtData);
        }
        return { failed: allFailed, passed: allPassed };
    }
    /**
     * Runs the script for each package in `packItems`
     */
    async runScripts(pkgRunManifest) {
        if (!pkgRunManifest) {
            throw new TypeError('(runScripts) "pkgRunManifest" arg is required');
        }
        const pkgRunManifestForEmit = {};
        for (const [pm, runManifests] of pkgRunManifest) {
            const pmId = this.pmIds.get(pm);
            if (!pmId) {
                /* istanbul ignore next */
                throw new error_1.PackageManagerIdError();
            }
            pkgRunManifestForEmit[pmId] = [...runManifests];
        }
        const totalScripts = [...pkgRunManifest].reduce((count, [, runManifests]) => count + runManifests.size, 0);
        await Promise.resolve();
        this.emit(event_1.Event.RUN_SCRIPTS_BEGIN, {
            manifest: pkgRunManifestForEmit,
            total: totalScripts,
        });
        const { bail } = this;
        let current = 0;
        const scripts = [];
        const results = [];
        BAIL: for await (const [pm, runManifests] of pkgRunManifest) {
            const pmId = this.pmIds.get(pm);
            if (!pmId) {
                /* istanbul ignore next */
                throw new error_1.PackageManagerIdError();
            }
            for await (const runManifest of runManifests) {
                let result;
                const { script } = runManifest;
                const { pkgName } = runManifest.packedPkg;
                scripts.push(script);
                this.emit(event_1.Event.RUN_SCRIPT_BEGIN, {
                    script,
                    pkgName,
                    total: totalScripts,
                    current,
                });
                try {
                    result = await pm.runScript(runManifest);
                    results.push(result);
                    if (result.error) {
                        debug('Script "%s" failed in package "%s": %O', script, pkgName, result);
                        this.emit(event_1.Event.RUN_SCRIPT_FAILED, {
                            ...result,
                            script,
                            error: result.error,
                            current: current++,
                            total: totalScripts,
                        });
                        if (bail) {
                            break BAIL;
                        }
                    }
                    else {
                        this.emit(event_1.Event.RUN_SCRIPT_OK, {
                            ...result,
                            script,
                            current,
                            total: totalScripts,
                        });
                    }
                }
                catch (err) {
                    throw new error_1.PackageManagerError(`Package manager "${pmId}" failed to run script "${script}`, pmId, err);
                }
            }
        }
        const failed = results.filter((result) => result.error).length;
        const passed = results.length - failed;
        this.emit(failed ? event_1.Event.RUN_SCRIPTS_FAILED : event_1.Event.RUN_SCRIPTS_OK, {
            results,
            manifest: pkgRunManifestForEmit,
            total: totalScripts,
            failed,
            passed,
        });
        return results;
    }
    /**
     * Converts install results to a {@linkcode PkgRunManifest} for {@linkcode runScripts}
     * @param installResults Results of {@linkcode install}
     * @returns Package managers and scripts to run for each unpacked package
     */
    buildPkgRunManifest(installResults) {
        const pkgRunManifest = new Map();
        for (const [pm, [{ packedPkgs }]] of installResults) {
            const runManifests = new Set();
            for (const packedPkg of packedPkgs) {
                for (const script of this.scripts) {
                    runManifests.add({
                        script,
                        packedPkg,
                    });
                }
            }
            pkgRunManifest.set(pm, runManifests);
        }
        return pkgRunManifest;
    }
    /**
     * Returns `true` if any of the scripts failed or any of the checks failed
     * @param results Results from {@linkcode smoke}
     * @returns Whether the results indicate a failure
     */
    isSmokeFailure({ scripts, checks }) {
        return (checks.failed.some((checkFailure) => checkFailure.severity === rules_1.CheckSeverities.ERROR) || scripts.some((runScriptResult) => runScriptResult.error));
    }
    /**
     * Pack, install, run checks (optionally), and run scripts (optionally)
     * @returns Results
     */
    async smoke() {
        // do not emit synchronously
        await Promise.resolve();
        this.emit(event_1.Event.SMOKE_BEGIN);
        try {
            // PACK
            const pkgInstallManifest = await this.pack();
            // INSTALL
            const installResults = await this.install(pkgInstallManifest);
            let ruleResults = { passed: [], failed: [] };
            let runScriptResults = [];
            // RUN CHECKS
            if (this.checks) {
                ruleResults = await this.runChecks(installResults);
            }
            // RUN SCRIPTS
            if (this.scripts.length) {
                const pkgRunManifest = this.buildPkgRunManifest(installResults);
                runScriptResults = await this.runScripts(pkgRunManifest);
            }
            // END
            const smokeResults = {
                scripts: runScriptResults,
                checks: ruleResults,
                opts: this.originalOpts,
            };
            if (this.isSmokeFailure(smokeResults)) {
                this.emit(event_1.Event.SMOKE_FAILED, new error_1.SmokeFailedError('🤮 Maurice!', { results: smokeResults }));
            }
            else {
                this.emit(event_1.Event.SMOKE_OK, smokeResults);
            }
            return smokeResults;
        }
        finally {
            await this.cleanup();
        }
    }
    /**
     * This is only here because it's a fair amount of work to mash the data into a format more suitable for display.
     *
     * This is used by the events {@linkcode SmokerEvent.InstallBegin}, {@linkcode SmokerEvent.InstallOk}, and {@linkcode SmokerEvent.PackOk}.
     * @param pkgInstallManifest What to install and with what package manager
     * @returns Something to be emitted
     */
    buildInstallEventData(pkgInstallManifest) {
        const uniquePkgs = new Set();
        const pmIds = new Set();
        const manifests = [];
        const additionalDeps = new Set();
        for (const [pm, manifest] of pkgInstallManifest) {
            const id = this.pmIds.get(pm);
            if (!id) {
                /* istanbul ignore next */
                throw new error_1.PackageManagerIdError();
            }
            pmIds.add(id);
            manifests.push(manifest);
            for (const { pkgName } of manifest.packedPkgs) {
                uniquePkgs.add(pkgName);
            }
            if (manifest.additionalDeps) {
                for (const dep of manifest.additionalDeps) {
                    additionalDeps.add(dep);
                }
            }
        }
        return {
            uniquePkgs: [...uniquePkgs],
            packageManagers: [...pmIds],
            additionalDeps: [...additionalDeps],
            manifests,
        };
    }
}
exports.Smoker = Smoker;
//# sourceMappingURL=smoker.js.map