"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.findDataDir = exports.readPackageJsonSync = exports.readPackageJson = exports.pickPackageVersion = exports.castArray = exports.normalizeStringArray = void 0;
/**
 * Gotta have a "util" module
 * @module
 */
const node_path_1 = __importDefault(require("node:path"));
const read_pkg_up_1 = __importDefault(require("read-pkg-up"));
const error_1 = require("./error");
/**
 * Trims all strings in an array and removes empty strings.
 * Returns empty array if input is falsy.
 */
function normalizeStringArray(value) {
    return value
        ? castArray(value)
            .map((item) => item.trim())
            .filter(Boolean)
        : [];
}
exports.normalizeStringArray = normalizeStringArray;
function castArray(value) {
    return (value ? (Array.isArray(value) ? value : [value]) : []).filter(Boolean);
}
exports.castArray = castArray;
/**
 * Regex string to match a package name.
 *
 * Used by {@linkcode PKG_NAME_REGEX} and {@linkcode PKG_NAME_WITH_SPEC_REGEX}.
 */
const PKG_NAME_REGEX_STR = '^(@[a-z0-9-~][a-z0-9-._~]*/)?[a-z0-9-~][a-z0-9-._~]*';
/**
 * Regex to match a package name without a spec
 */
const PKG_NAME_REGEX = new RegExp(`${PKG_NAME_REGEX_STR}$`);
/**
 * Regex to match a package name with a spec.
 *
 * @remarks This does not attempt to validate a semver string, though it could.
 * If it did, it'd also need to allow any valid package tag.  I'm not sure what
 * the latter is, but the former can be found on
 * {@link https://stackoverflow.com/a/72900791|StackOverflow}.
 */
const PKG_NAME_WITH_SPEC_REGEX = new RegExp(`${PKG_NAME_REGEX_STR}@.+$`);
/**
 * Fields in `package.json` that might have a dependency we want to install as
 * an isolated package to help run smoke tests.
 *
 * @remarks Order is important; changing this should be a breaking change
 */
const DEP_FIELDS = [
    'devDependencies',
    'dependencies',
    'optionalDependencies',
    'peerDependencies',
];
/**
 * Try to pick a version for a package to install.
 *
 * Given an `installable` which is both a) a valid npm package name and b) has
 * no version specifier, determine the version to install.
 *
 * If the `package.json` within `cwd` contains the package of the same name, we
 * will use that version; otherwise we will use the `latest` tag. If
 * `installable` is not a package name at all, it passes thru verbatim.
 *
 * @param installable The `thing` in `npm install <thing>`
 * @param cwd Where the command would be run
 */
async function pickPackageVersion(installable, cwd = process.cwd()) {
    if (PKG_NAME_WITH_SPEC_REGEX.test(installable)) {
        // we were given a package name with a version spec. just use it
        return installable;
    }
    if (PKG_NAME_REGEX.test(installable)) {
        // we were given a package name, no version.
        // try to see if it's in the package.json
        const pkgName = installable;
        const { packageJson } = (await readPackageJson({ cwd })) ?? {};
        if (packageJson) {
            for (const field of DEP_FIELDS) {
                const deps = packageJson[field];
                if (deps && pkgName in deps) {
                    return `${pkgName}@${deps[pkgName]}`;
                }
            }
        }
        return `${pkgName}@latest`;
    }
    // could be a path or url
    return installable;
}
exports.pickPackageVersion = pickPackageVersion;
async function readPackageJson({ cwd, normalize, strict, } = {}) {
    cwd ??= process.cwd();
    if (readPackageJson.cache.has({ cwd, normalize })) {
        return readPackageJson.cache.get({ cwd, normalize });
    }
    try {
        const result = await (0, read_pkg_up_1.default)({ cwd, normalize });
        if (!result && strict) {
            throw new error_1.MissingPackageJsonError(`Could not find package.json from ${cwd}`, cwd);
        }
        readPackageJson.cache.set({ cwd, normalize }, result);
        return result;
    }
    catch (err) {
        throw new error_1.UnreadablePackageJsonError(`Could not read package.json from ${cwd}`, cwd, err);
    }
}
exports.readPackageJson = readPackageJson;
readPackageJson.cache = new Map();
function readPackageJsonSync({ cwd, normalize, strict, } = {}) {
    cwd ??= process.cwd();
    if (readPackageJsonSync.cache.has({ cwd, normalize })) {
        return readPackageJsonSync.cache.get({ cwd, normalize });
    }
    try {
        const result = read_pkg_up_1.default.sync({ cwd, normalize });
        if (!result && strict) {
            throw new error_1.MissingPackageJsonError(`Could not find package.json from ${cwd}`, cwd);
        }
        readPackageJsonSync.cache.set({ cwd, normalize }, result);
        return result;
    }
    catch (err) {
        throw new error_1.UnreadablePackageJsonError(`Could not read package.json from ${cwd}`, cwd, err);
    }
}
exports.readPackageJsonSync = readPackageJsonSync;
readPackageJsonSync.cache = new Map();
let dataDir;
async function findDataDir() {
    if (dataDir) {
        return dataDir;
    }
    const { path: packagePath } = (await readPackageJson({
        cwd: __dirname,
    }));
    const root = node_path_1.default.dirname(packagePath);
    dataDir = node_path_1.default.join(root, 'data');
    return dataDir;
}
exports.findDataDir = findDataDir;
//# sourceMappingURL=util.js.map